"""
"""
from behave import *

import behave_restful._lang_imp.request_builder as _builder

@given('a request url {url}')
def step_impl(context, url):
    _builder.set_url(context, url)


@given('a request json payload')
def step_impl(context):
    _builder.set_json_payload(context, context.text)


@given('request headers')
def step_impl(context):
    _builder.set_request_headers(context, context.table)


@given('request parameters')
def step_impl(context):
    _builder.set_request_params(context, context.table)


"""
Implements the Behave Restful application object which initializes the global
execution context providing access to directories and other services to 
simplify testing.  
"""
import os
import shutil

import requests

import behave_restful._definitions as _defs
import behave_restful._hooks as _hooks

from behave_restful._hooks import (
    BEFORE_ALL, 
    AFTER_ALL, 
    BEFORE_FEATURE, 
    AFTER_FEATURE, 
    BEFORE_SCENARIO, 
    AFTER_SCENARIO,
    BEFORE_STEP,
    AFTER_STEP,
    BEFORE_TAG,
    AFTER_TAG
)

class BehaveRestfulApp(object):
    """
    Behave Restful application class used to initialize the execution context
    for a test run.
    """

    def initialize_context(self, context, test_dir):
        """
        Initializes the provide context object exposing through it relevant
        directories and services. 

        :param context:
            A behave context object.
        :param test_dir:
            Directory where the tests are located. This is, usually, the 
            features folder of the test project.
        """
        context.test_dir = test_dir
        context.working_dir = os.getcwd()
        context.default_session = requests
        context.session = context.default_session
        _defs.DefinitionInitializer().initialize(context)
        _hooks.EnvironmentHookInitializer().initialize(context)



    
class BrInitApp(object):
    """
    Application class for the br-init command that creates a new Behave Restful
    project.
    """
    def __init__(self):
        self.project_dir = os.path.join(os.getcwd(), 'features')
        behave_restful_root = os.path.dirname(os.path.abspath(__file__))
        self.src_dir = os.path.join(behave_restful_root, '_project')


    def init_project(self, project_dir=None):
        """
        Creates a new project by copying the default template files to the
        specified destination, or the features folder in the current working
        directory.

        The destination folder cannot exist before running this application.
        """
        self.project_dir = project_dir or self.project_dir
        msg = 'Creating project at: {d}'.format(d=self.project_dir)
        print(msg)
        self._check_project_dir()
        self._copy_project_files()


    def _check_project_dir(self):
        if self._exists(self.project_dir):
            raise ProjectInitError('Target directory already exists')

    
    def _copy_project_files(self):
        self._copy_tree(self.src_dir, self.project_dir)


    def _exists(self, path):
        return os.path.exists(path)


    def _copy_tree(self, src, dst):
        shutil.copytree(src, dst)



class ProjectInitError(Exception):
    """
    """
    def __init__(self, reason):
        super(ProjectInitError, self).__init__()
        self.reason = reason


    def __repr__(self):
        return 'ProjectInitError({r})'.format(r=self.reason)



"""
Provides properties with information about ``behave_restful``.
"""

project = 'behave-restful'
version = '0.4'
release = '{version}.4'.format(version=version)
description = 'Implements Gherking language for REST services.'
copyright = '2017 Abantos'
author = 'Isaac Rodriguez'
author_email = 'oss.abantos@outlook.com'
url = 'https://github.com/behave-restful/behave-restful'
license = 'MIT'
keywords = 'rest bdd behave gherkin test automation testing'

classifiers = [
    'Development Status :: 1 - Planning',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Software Development :: Testing :: Acceptance',
    'Topic :: Software Development :: Testing :: BDD',
]



"""
This module imports all the REST language implementation steps into the project
so you can use them to write the tests. You can implement additional steps in
this module, or better yet, create additional step files in this folder for 
your custom steps.
"""
from behave_restful.lang import *



"""
Exposes functions to provide a command line interface to Behave Restful.
"""
import sys

import behave_restful.app as app

def behave_restful_init():
    try:
        target_dir = sys.argv[1] if len(sys.argv) > 1 else None
        application = app.BrInitApp()
        application.init_project(target_dir)
    except app.ProjectInitError as err:
        print(err.reason)
        sys.exit(2)
    except Exception as err:
        msg = 'Error creating project {e}'.format(e=err)
        print(msg)
        sys.exit(1)



"""
"""
import json

from assertpy import assert_that, fail
from jsonpath import jsonpath
import jsonschema

from behave_restful.xpy import HTTPStatus

def response_status_is(response, expected_status):
    """
    """
    expected_status = _as_numeric_status(expected_status)
    actual_status = response.status_code
    assert_that(actual_status).is_equal_to(expected_status)


def response_json_matches(response, schema_str):
    """
    """
    schema = json.loads(schema_str)
    json_body = response.json()
    _validate_with_schema(json_body, schema)


def response_json_matches_defined_schema(context, schema_id):
    """
    """
    schema_id = context.vars.resolve(schema_id)
    schema = context.schemas.get(schema_id)
    json_body = context.response.json()
    _validate_with_schema(json_body, schema)


def response_json_at_path_is_equal_to(response, json_path, value):
    """
    """
    values = _get_values(response.json(), json_path)
    [assert_that(actual_value).is_equal_to(eval(value)) for actual_value in values]
    

def response_json_at_path_is_not_equal_to(response, json_path, value):
    """
    """
    values = _get_values(response.json(), json_path)
    [assert_that(actual_value).is_not_equal_to(eval(value)) for actual_value in values]


def response_json_at_path_starts_with(response, json_path, value):
    """
    """
    values = _get_values(response.json(), json_path)
    [assert_that(actual_value).starts_with(eval(value)) for actual_value in values]


def response_json_at_path_ends_with(response, json_path, value):
    """
    """
    values = _get_values(response.json(), json_path)
    [assert_that(actual_value).ends_with(eval(value)) for actual_value in values]


def response_json_at_path_contains(response, json_path, value):
    """
    """
    values = _get_values(response.json(), json_path)
    [assert_that(actual_value).contains(eval(value)) for actual_value in values]


def response_json_at_path_does_not_contain(response, json_path, value):
    """
    """
    values = _get_values(response.json(), json_path)
    [assert_that(actual_value).does_not_contain(eval(value)) for actual_value in values]
    

def response_json_at_path_is_null(response, json_path):
    """
    """
    values = _get_values(response.json(), json_path)
    [assert_that(actual_value).is_none() for actual_value in values]


def response_json_at_path_is_not_null(response, json_path):
    """
    """
    values = _get_values(response.json(), json_path)
    [assert_that(actual_value).is_not_none() for actual_value in values]


def response_json_at_path_is_true(response, json_path):
    """
    """
    values = _get_values(response.json(), json_path)
    [assert_that(actual_value).is_true() for actual_value in values]


def response_json_at_path_is_false(response, json_path):
    """
    """
    values = _get_values(response.json(), json_path)
    [assert_that(actual_value).is_false() for actual_value in values]


def response_header_is_equal_to(response, header_name, value):
    """
    """
    actual_value = _get_header_values(response.headers, header_name)
    assert_that(actual_value).is_equal_to(value)


def _as_numeric_status(status):
    status = status.replace(' ', '_')
    numeric_status = getattr(HTTPStatus, status.upper(), None)
    if not numeric_status:
        numeric_status = int(status)
    return numeric_status


def _get_values(json_body, json_path):
    results = jsonpath(json_body, json_path)
    if not results: fail('Match not found at <{path}> for <{body}>'.format(path=json_path, body=json_body))
    return results

def _get_header_values(headers, header_name):
    result = headers.get(header_name)
    if not result: fail('Match not found at <{headers}> for <{header_name}>'.format(headers=headers, header_name=header_name))
    return result

def _validate_with_schema(json_body, schema)  :
    jsonschema.validate(json_body, schema)




"""
All python modules in this folder are loaded by Behave Restful and the
implemented hooks are registered to be invoked at the correct time. You can
uncomment the hook you need and implement it or create a new hooks file in this
folder using this one as a base.
"""

# def before_all(context):
#     pass


# def after_all(context):
#     pass
    

# def before_feature(context, feature):
#     pass


# def after_feature(context, feature):
#     pass


# def before_scenario(context, scenario):
#     pass


# def after_scenario(context, scenario):
#     pass


# def before_step(context, step):
#     pass


# def after_step(context, step):
#     pass


# def before_tag(context, tag):
#     pass


# def after_tag(context, tag):
#     pass


"""
"""
import json
import os.path

import jsonref
import yaml

def add_json_schemas(schemas, context):
    """
    """
    if not hasattr(context, 'schemas') or not context.schemas:
        context.schemas = {}
    context.schemas.update(schemas)


def add_openapi_schemas(openapi_filename, context):
    abs_path = os.path.abspath(openapi_filename)
    root, filename = os.path.split(openapi_filename)
    loader = ReferenceFileLoader(root)
    reader = OpenApiReader(filename, loader)
    add_json_schemas(reader.schemas, context)


class OpenApiReader:
    def __init__(self, filename, loader):
        self._filename = filename
        self._loader = loader
        self._content = None


    @property
    def content(self):
        if not self._content:
            self._content = self._load()
        return self._content


    @property
    def schemas(self):
        return self.content.get('components').get('schemas')


    def _load(self):
        content = self._loader(self._filename)
        content = jsonref.JsonRef.replace_refs(content, loader=self._loader)
        return content





class ReferenceFileLoader:
    def __init__(self, root_dir):
        self.root_dir = root_dir

        
    def __call__(self, ref_file):
        full_name = os.path.join(self.root_dir, ref_file)
        reader = get_reader(ref_file)
        return self._load(reader, full_name)


    def _load(self, reader, filename):
        with open(filename) as fin:
            return reader.load(fin)


YAML_EXT = '.yaml'

def get_reader(filename):
    name, ext = os.path.splitext(filename)
    if ext == YAML_EXT:
        return yaml_reader
    return json


class _YamlReader:

    def load(self, fin):
        return yaml.load(fin, Loader=yaml.Loader)

yaml_reader = _YamlReader()



import os

import behave_restful.app as br_app


def before_all(context):
    this_directory = os.path.abspath(os.path.dirname(__file__))
    br_app.BehaveRestfulApp().initialize_context(context, this_directory)
    context.hooks.invoke(br_app.BEFORE_ALL, context)


def after_all(context):
    context.hooks.invoke(br_app.AFTER_ALL, context)


def before_feature(context, feature):
    context.hooks.invoke(br_app.BEFORE_FEATURE, context, feature)


def after_feature(context, feature):
    context.hooks.invoke(br_app.AFTER_FEATURE, context, feature)


def before_scenario(context, scenario):
    context.hooks.invoke(br_app.BEFORE_SCENARIO, context, scenario)


def after_scenario(context, scenario):
    context.hooks.invoke(br_app.AFTER_SCENARIO, context, scenario)


def before_step(context, step):
    context.hooks.invoke(br_app.BEFORE_STEP, context, step)


def after_step(context, step):
    context.hooks.invoke(br_app.AFTER_STEP, context, step)


def before_tag(context, tag):
    context.hooks.invoke(br_app.BEFORE_TAG, context, tag)


def after_tag(context, tag):
    context.hooks.invoke(br_app.AFTER_TAG, context, tag)


"""
"""
from behave import *

import behave_restful._lang_imp.response_validator as _validate

@then('the response status is {status}')
def step_impl(context, status):
    _validate.response_status_is(context.response, status)


@then('the response json matches')
def step_impl(context):
    schema = context.vars.resolve(context.text)
    _validate.response_json_matches(context.response, schema)


@then('the response json matches defined schema {schema_id}')
def step_impl(context, schema_id):
    _validate.response_json_matches_defined_schema(context, schema_id)


@then('the response json at {json_path} is equal to {value_str}')
def step_impl(context, json_path, value_str):
    _invoke_match(_validate.response_json_at_path_is_equal_to, context, json_path, value_str)


@then('the response json at {json_path} is not equal to {value_str}')
def step_impl(context, json_path, value_str):
    _invoke_match(_validate.response_json_at_path_is_not_equal_to, context, json_path, value_str)


@then('the response json at {json_path} starts with {value_str}')
def step_impl(context, json_path, value_str):
    _invoke_match(_validate.response_json_at_path_starts_with, context, json_path, value_str)


@then('the response json at {json_path} ends with {value_str}')
def step_impl(context, json_path, value_str):
    _invoke_match(_validate.response_json_at_path_ends_with, context, json_path, value_str)


@then('the response json at {json_path} contains {value_str}')
def step_impl(context, json_path, value_str):
    _invoke_match(_validate.response_json_at_path_contains, context, json_path, value_str)


@then('the response json at {json_path} does not contain {value_str}')
def step_impl(context, json_path, value_str):
    _invoke_match(_validate.response_json_at_path_does_not_contain, context, json_path, value_str)


@then('the response json at {json_path} is null')
def step_impl(context, json_path):
    _check_value_with(_validate.response_json_at_path_is_null, context, json_path)


@then('the response json at {json_path} is not null')
def step_impl(context, json_path):
    _check_value_with(_validate.response_json_at_path_is_not_null, context, json_path)


@then('the response json at {json_path} is true')
def step_impl(context, json_path):
    _check_value_with(_validate.response_json_at_path_is_true, context, json_path)


@then('the response json at {json_path} is false')
def step_impl(context, json_path):
    _check_value_with(_validate.response_json_at_path_is_false, context, json_path)


@then('the response header {header} is equal to {value}')
def step_impl(context, header, value):
    _invoke_match_header(_validate.response_header_is_equal_to, context, header, value)


def _invoke_match(func, context, json_path, value_str):
    json_path = context.vars.resolve(json_path)
    value_str = context.vars.resolve(value_str)
    func(context.response, json_path, value_str)


def _invoke_match_header(func, context, header, value_str):
    header = context.vars.resolve(header)
    value_str = context.vars.resolve(value_str)
    func(context.response, header, value_str)


def _check_value_with(func, context, json_path):
    json_path = context.vars.resolve(json_path)
    func(context.response, json_path)



"""
"""
import json

def set_url(context, url):
    """
    """
    resolved_url = context.vars.resolve(url)
    context.request_url = resolved_url


def set_json_payload(context, payload):
    """
    """
    resolved_payload = context.vars.resolve(payload)
    context.request_json_payload = json.loads(resolved_payload)


def set_request_params(context, params):
    """
    """
    resolve = context.vars.resolve
    resolved_params = {resolve(param['param']): resolve(param['value']) for param in params}
    context.request_params = resolved_params

def set_request_headers(context, headers):
    """
    """
    resolve = context.vars.resolve
    resolved_headers = {resolve(header['param']): resolve(header['value']) for header in headers}
    context.request_headers = resolved_headers


"""
This module exposes functionality to handle definition files in the application.
"""
import os

import behave_restful._errors as _errors
import behave_restful._utils as _utils
import behave_restful.xpy as xpy

DEFINITIONS_DIR = 'definitions'
DEFINITION_KEY = 'definition'
class DefinitionInitializer(object):
    """
    Initializes definitions into the execution context.
    """
    
    def initialize(self, context):
        """
        Initializes the specified definition, if exists, into the execution
        context.

        :param context:
            Context object to initialize.
        """
        context.vars = VarsManager()
        self._add_to_search_path(context.test_dir, DEFINITIONS_DIR)
        if DEFINITION_KEY in context.config.userdata:
            self._initialize_definition(context)


    def _add_to_search_path(self, *path_tokens):
        _utils.add_search_path(*path_tokens)


    def _initialize_definition(self, context):
        definition_name = context.config.userdata[DEFINITION_KEY]
        definition_module = self._load_definition(definition_name)
        if not definition_module: raise DefinitionNotFoundError(definition_name)
        definition_module.initialize_definition(context)


    def _load_definition(self, definition_name):
        return _utils.load_module(definition_name)





VAR_PREFIX = '${'
VAR_SUFFIX = '}'

_NO_NEXT_VAR = None

class VarsManager(object):
    """
    Manages access and resolution of variables defined in definition files or
    in the environment.
    """
    def __init__(self):
        self._definitions = {}


    def add(self, var_name, var_value):
        """
        Adds a new variable. If the variable already existed, the value is
        overwritten.

        :param str var_name: 
            Name of the variable to add.
        :param var_value: 
            Value to set the variable to.
        :type var_value:
            any
        """
        self._definitions[var_name] = var_value


    def add_vars(self, definitions):
        """
        Adds all the variables specified in the definition. 
        
        The definition parameter must be a dictionary where the keys are used as 
        variable names and the values as the values for the variables. Existing 
        variables will be replaced.

        :param dict definitions:
            Dictionary containing the variable definitions.
        """
        self._definitions.update(definitions)


    def get(self, var_name, def_value=None):
        """
        Returns the resolved value of a variable.

        :param str var_name:
            Name of the variable to retrieve value.
        :param def_value:
            Optional default value to be returned if the variable is not 
            defined.

        :return:
            The resolved value of the variable.
        """
        environment_value = os.environ.get(var_name)
        value = environment_value or self._definitions.get(var_name, def_value)
        return self.resolve(value) if xpy.is_string(value) else value
        

    def resolve(self, to_resolve):
        """
        Resolves a string containing embedded variables.

        :param str to_resolve:
            String in which we want to resolve the embedded variables.
        
        :return:
            The string with embedded variables resolved or the same string if
            variable cannot be resolved or the string does not contain embedded
            variables.
        :rtype:
            str

        The following code sample shows how this function works:

        ..  code-block:: python

            variable_definitions = {
                'INT_VAR': 1,
                'STR_VAR': 'a string',
                'EMBEDDED_VARS': 'contains ${INT_VAR} and ${STR_VAR},
            }

            manager = VarsManager()
            manager.add_vars(variable_definitions)

            manager.resolve('contains ${INT_VAR}')
            # returns 'contains 1'

            manager.resolve('contains ${STR_VAR}')
            # returns 'contains a string'

            manager.resolve('contains ${INT_VAR} and ${STR_VAR}')
            # returns 'contains 1 and a string'

            manager.resolve('this ${EMBEDDED_VARS}')
            # returns 'this contains 1 and a string'

            manager.resolve('this is ${UNDEFINED}')
            # returns 'this is ${UNDEFINED}'
        """
        resolved = to_resolve
        next_var_name = self._find_var_name_in(resolved)
        while next_var_name:
            resolved, next_var_name = self._resolve_in(resolved, next_var_name)
        return resolved


    def _resolve_in(self, to_resolve, var_name):
        value = self.get(var_name)
        return self._do_resolve_in(to_resolve, var_name, value) if value is not None else (to_resolve, _NO_NEXT_VAR)


    def _do_resolve_in(self, to_resolve, var_name, value):
        resolved = self._replace_with(to_resolve, var_name, value)
        next_var_name = self._find_var_name_in(resolved)
        return resolved, next_var_name



    def _find_var_name_in(self, to_resolve):
        prefix_location = to_resolve.find(VAR_PREFIX)
        if self._prefix_found(prefix_location):
            var_starts = prefix_location + len(VAR_PREFIX)
            var_ends = to_resolve.find(VAR_SUFFIX, var_starts)
            return to_resolve[var_starts:var_ends]


    def _replace_with(self, to_resolve, var_name, value):
        token = ''.join([VAR_PREFIX, var_name, VAR_SUFFIX])
        return to_resolve.replace(token, str(value))

    def _prefix_found(self, location):
        return location >= 0



class DefinitionNotFoundError(_errors.BehaveRestfulException):
    """
    """
    def __init__(self, definition_name):
        self.definition_name = definition_name


    def __repr__(self):
        return "DefinitionNotFoundError(definition_name='{dn}')".format(dn=self.definition_name)


    



"""
This module is loaded when 'sample' is passed as the definition to Behave Restful.
The variables here defined can then be used for testing. Definitions are used
to define variables with values matching for different testing environments.
"""

vars = {
    'BASE_URL': 'https://api.github.com',
}

def initialize_definition(context):
    context.vars.add_vars(vars)



"""
Setup for behave-restful
"""
from setuptools import setup, find_packages
from os import path

from m2r2 import parse_from_file

import behave_restful.about as about

readme_file = path.join(path.dirname(path.abspath(__file__)), 'README.md')
readme = parse_from_file(readme_file)

def _read_dependencies():
    requirements_file = 'dependencies.txt'
    with open(requirements_file) as fin:
        return [line.strip() for line in fin if line]

packages = find_packages()
requirements = _read_dependencies()
entry_points = {
    'console_scripts': [
        'br-init = behave_restful.cli:behave_restful_init'
    ]
}

package_data = {
    'behave_restful': ['./_project/**/*']
}

setup(
    name=about.project,
    version=about.release,
    description=about.description,
    long_description=readme,
    author=about.author,
    author_email=about.author_email,
    url=about.url,
    license=about.license,
    keywords=about.keywords,
    classifiers=about.classifiers,
    packages=packages,
    install_requires=requirements,
    entry_points=entry_points,
    package_data=package_data
)



"""
"""
from behave import *

import behave_restful._lang_imp.request_invoker as _invoker

@when('the request sends {method}')
def step_impl(context, method):
    invoker_function_name = 'send_' + method.lower()
    request_method = getattr(_invoker, invoker_function_name)
    request_method(context)